/* 
 * xMedia
 * Serial Interface Engine
 * 
 *  General Data Frame:
 * ---------------------------------------------------------------------------------
 * | Control 16bit | Command 8bit	| Length 8bit	|  Data	variable |  CRC8 8bit  |
 *  --------------------------------------------------------------------------------
 * 
 * The packages are sent using the SLIP (http://tools.ietf.org/html/rfc1055)
 * 
 * 
 * Control:
 *  ---------------------------------------------------------------
 * |ANS|D14|D13|D12|D11|D10|D09|D08|D07|D06|D05|D04|D03|D02|D01|D00|
 *  ---------------------------------------------------------------
 * 
 * ANS 	... is answer to previous command
 * D00-D14 ... package identifier
 * 
 * 
 * Command:
 *  -------------------------------
 * |R07|R06|R05|R04|R03|R02|R01|R00|
 *  -------------------------------
 * 
 * R00-R7	... Register descriptions
 * 
 * 
 * Length:
 *  -------------------------------------
 * |				Length				|
 *  -------------------------------------
 * 
 * Length	... Length of payload data (in bytes)
 * 
 * Data:
 *  -------------------------------------
 * |		Data (Variable length)		|
 *  -------------------------------------
 * 
 * Data	... Payload data (length specified in Length field) 
 * 
 * 
 * CRC:
 *  -------------------------------------
 * |			CRC (8bit)				|
 *  -------------------------------------
 * 
 * CRC	... Checksum for the WHOLE data package (optional)
 * 	    write this field to 0x00 if not used 
 * 
 * 
 * Limitations:
 * The maximum length of the allowed payload data is limited to
 * SIE_MAX_PACKET_SIZE - 5 byte or to a maximum of 250 byte.
 * Background: This is done to increase the performance of the SIE
 * on 8bit microprocessors (almost double the speed of the data)
 * reception and submission.
 */


#include "sie.h"
#include <avr/interrupt.h>

#define SIE_END             0300    /* indicates end of packet */
#define SIE_ESC             0333    /* indicates byte stuffing */
#define SIE_ESC_END         0334    /* ESC ESC_END means END data byte */
#define SIE_ESC_ESC         0335    /* ESC ESC_ESC means ESC data byte */

unsigned char sie_packet_id = 0;

unsigned char sie_created = 0; 
unsigned char sie_framestart[SIE_COUNT];
unsigned char sie_receive_temp[SIE_COUNT][SIE_MAX_PACKET_LENGTH];
unsigned char sie_count[SIE_COUNT];
unsigned char sie_state[SIE_COUNT];

void (*sie_receive_callback[SIE_COUNT])(siePacket);
void (*sie_error_callback[SIE_COUNT])(unsigned char);
void (*sie_send_callback[SIE_COUNT])(unsigned char);


void sieInit()
{

}

unsigned char sieCrcCreate(sieRawPacket data, unsigned char length)
{
	//TODO ADD CRC generation
	return 0;
}

unsigned char sieCrcCheck(siePacket data)
{
	
	return 1;
}


unsigned char sieCreate(void (*receive_callback)(siePacket), void (*send_callback)(unsigned char), void (*error_callback)(unsigned char))
{
     sie_framestart[sie_created] = 0;
     sie_receive_callback[sie_created] = receive_callback;
     sie_error_callback[sie_created] = error_callback;
     sie_send_callback[sie_created] = send_callback;
     
     return sie_created++;
}

unsigned char siePackData(unsigned char sie_id)
{
	siePacket packet;
	int i = 0;
	
	packet.pid 	= sie_receive_temp[sie_id][0] << 8 | sie_receive_temp[sie_id][1];
	packet.command 	= sie_receive_temp[sie_id][2];
	packet.length	= sie_receive_temp[sie_id][3];
	
	for(i = 0; i < packet.length && i < SIE_MAX_PACKET_SIZE; i++)
	{
		packet.data[i] = sie_receive_temp[sie_id][i + 4];
	}
	
	packet.crc		= sie_receive_temp[sie_id][i+4];
	
	
	if(!sieCrcCheck(packet))
	{
		(sie_error_callback[sie_id])(SIE_CRC_CHECK_FAILED);		
	}
	else
	{
		(sie_receive_callback[sie_id])(packet);		
	}
}

unsigned char siePackRaw(unsigned char sie_id, sieRawPacket buffer, unsigned char pid, unsigned char command, unsigned char length, unsigned char *data)
{
	int i = 0;

	buffer[0] = pid >> 8; //MSB
	buffer[1] = pid & 0x00FF; // LSB
	buffer[2] = command;
	
	
	for(i = 0; i < length && i < SIE_MAX_PACKET_SIZE; i++)
	{
		buffer[i+4] = data[i];
	}
	
	buffer[3] = i;
	
	buffer[4+i] = sieCrcCreate(buffer, i+3);
	
	return i+5;
}


/* ATOMIC OPERATION!!!! */
void sieSubmitRaw(unsigned char sie_id, sieRawPacket raw, unsigned char length)
{
	int i = 0;

	/* Transmission MUST NOT be disturbed */ 
	cli();	
	for(i = 0; i < length; i++)
	{
		if(raw[i] == SIE_END)
		{
			(sie_send_callback[sie_id])(SIE_ESC);
			(sie_send_callback[sie_id])(SIE_ESC_END);	
		}
		else if(raw[i] == SIE_ESC)
		{
			(sie_send_callback[sie_id])(SIE_ESC);
			(sie_send_callback[sie_id])(SIE_ESC_ESC);				
		}
		else
		{
			(sie_send_callback[sie_id])(raw[i]);			
		}
	}
	sei();	
}

void sieSubmitPacket(unsigned char sie_id, siePacket data)
{
	sieRawPacket buffer;
	unsigned char plength = 0;
	plength = siePackRaw(sie_id, buffer, data.pid, data.command, data.length, data.data);
	sieSubmitRaw(sie_id, buffer, plength);
	
}

void sieSubmit(unsigned char sie_id, unsigned char command, unsigned char length, unsigned char *data)
{
	sieRawPacket buffer;
	unsigned char plength = 0;
	plength = siePackRaw(sie_id, buffer, sie_packet_id++, command, length, data);
	sieSubmitRaw(sie_id, buffer, plength);	
}

void sieRespond(unsigned char sie_id, unsigned char pid, unsigned char command, unsigned char length, unsigned char *data)
{
	sieRawPacket buffer;
	unsigned char plength = 0;
	plength = siePackRaw(sie_id, buffer, pid | (1<<15), command, length, data);
	sieSubmitRaw(sie_id, buffer, plength);	
}


void sieReceive(unsigned char sie_id, unsigned char data)
{ 
	 switch (sie_state[sie_id])
	 {
			case SIE_IDLE:
			{
			     if(data == SIE_END) //End of invalid transmission detected
			     {
						(sie_error_callback[sie_id])(SIE_ILLEGAL_FRAMESTART);
						
						int i = 0;
						for(i = 0; i < SIE_MAX_PACKET_LENGTH; i++)
						{
							sie_receive_temp[sie_id][i] = 0;
						}
						sie_framestart[sie_id] = 0;						

			            sie_state[sie_id] = SIE_RECEIVING;             
			     }
			     
			     break;
			}
			case SIE_RECEIVING:
			{
			     if(data == SIE_ESC)
			     {
			             sie_state[sie_id] = SIE_ESCAPING;
			     }
			     else if(data == SIE_END)
			     {
						siePackData(sie_id);
						
						int i = 0;
						for(i = 0; i < SIE_MAX_PACKET_LENGTH; i++)
						{
							sie_receive_temp[sie_id][i] = 0;
						}
						sie_framestart[sie_id] = 0;
						
						sie_state[sie_id] = SIE_RECEIVING; 
				 }
			     else
			     {
			             sie_receive_temp[sie_id][sie_framestart[sie_id]++] = data;
			    		 
			             sie_state[sie_id] = SIE_RECEIVING;       
			     }
			     
			     break;
			}
			case SIE_ESCAPING:
			{
				if(data == SIE_ESC_END)
				{
					sie_receive_temp[sie_id][sie_framestart[sie_id]++] = SIE_END;
					sie_state[sie_id] = SIE_RECEIVING;
				}
				else if(data == SIE_ESC_ESC)
				{
					sie_receive_temp[sie_id][sie_framestart[sie_id]++] = SIE_END;
					sie_state[sie_id] = SIE_RECEIVING;
				}
				else
				{
					(sie_error_callback[sie_id])(SIE_ILLEGAL_ESCAPE_SEQUENCE);
					
					sie_state[sie_id] = SIE_IDLE;
				}
			}               
	 }     
}
